#include "studio.h"
#include "codec_studio.h"
#include "bootdsp.h"
#include "fpga.h"

#define DELAY(time) snd_delay(time)

/* ADAT sync type. See studio.c */
extern int adat_sync[SND_CARDS];
/* Optional Sync Backplate Card installed. See studio.c */
extern int syncio[SND_CARDS];

/* WARNING. It is safe to make sure a litte delay -snd_delay(1)
 * seems enough- among DSP accesses within a burst. The reason
 * may be the DSP, the PCI bus or both.
 */

/**************************************************************/
static void snd_studio_codec_proc_read(
	snd_info_buffer_t	*buffer,
	void			*private_data)
{
	codec_studio_t	*codec = (codec_studio_t *) private_data;
	unsigned long	flags;
	unsigned	idx;

	snd_iprintf(buffer, "Sonorus Studi/o Codec #%d\n", codec->icodec);
	
	/* The lock assures meter readings are all form the same block */
	spin_lock_irqsave(&codec->access_lock, flags);
	snd_iprintf(buffer, "Playback peak meters (1-8)\t:");
	for (idx=0; idx<8; idx++)
		snd_iprintf(buffer, " 0x%0x", codec->status.peak_playback[idx]);
	snd_iprintf(buffer, "\n");
	snd_iprintf(buffer, "Playback peak meters (9-16)\t:");
	for (idx=8; idx<16; idx++)
		snd_iprintf(buffer, " 0x%0x", codec->status.peak_playback[idx]);
	snd_iprintf(buffer, "\n");
	snd_iprintf(buffer, "Record peak meters (1-8)\t:");
	for (idx=0; idx<8; idx++)
		snd_iprintf(buffer, " 0x%0x", codec->status.peak_record[idx]);
	snd_iprintf(buffer, "\n");
	snd_iprintf(buffer, "Record peak meters (9-16)\t:");
	for (idx=8; idx<16; idx++)
		snd_iprintf(buffer, " 0x%0x", codec->status.peak_record[idx]);
	snd_iprintf(buffer, "\n");
	snd_iprintf(buffer, "Monitor peak meters\t\t:");
	for (idx=0; idx<2; idx++)
		snd_iprintf(buffer, " 0x%0x", codec->status.peak_monitor[idx]);
	snd_iprintf(buffer, "\n");
	spin_unlock_irqrestore(&codec->access_lock, flags);

	if (codec->status.iostat & NO_ADAT_SYNC)
		snd_iprintf(buffer, "No ADAT sync\n");
	if (codec->status.iostat & INPUT_A_DATA_ERROR)
		snd_iprintf(buffer, "Input A Data Error\n");
	if (codec->status.iostat & INPUT_B_DATA_ERROR)
		snd_iprintf(buffer, "Input B Data Error\n");
	snd_iprintf(buffer, "\n");
	
	snd_iprintf(buffer, "Max. Retries due to DSP busy\t: %lu,\tTotal : %lu\n",
		codec->stat.nretries_dsp_busy,
		codec->stat.sum_retries_dsp_busy);
	snd_iprintf(buffer, "Max. Retries due to DSP ready\t: %lu,\tTotal : %lu\n",
		codec->stat.nretries_dsp_ready,
		codec->stat.sum_retries_dsp_ready);
	snd_iprintf(buffer, "Max. Retries due to not HTRQ\t: %lu,\tTotal : %lu\n",
		codec->stat.nretries_not_htrq,
		codec->stat.sum_retries_not_htrq);
	snd_iprintf(buffer, "Max. Retries due to not HRRQ\t: %lu,\tTotal : %lu\n",
		codec->stat.nretries_not_hrrq,
		codec->stat.sum_retries_not_hrrq);
	snd_iprintf(buffer, "Max. Retries due to not TRDY\t: %lu,\tTotal : %lu\n",
		codec->stat.nretries_not_trdy,
		codec->stat.sum_retries_not_trdy);
	snd_iprintf(buffer, "Max. Retries due to not run\t: %lu,\tTotal : %lu\n",
		codec->stat.nretries_not_run,
		codec->stat.sum_retries_not_run);
	snd_iprintf(buffer, "Max. Retries due to not idle\t: %lu,\tTotal : %lu\n",
		codec->stat.nretries_not_idle,
		codec->stat.sum_retries_not_idle);
	snd_iprintf(buffer, "Accumulated wait time (uS)\t: %lu\n",
		codec->stat.sum_wait_uS);
}

static int snd_codec_studio_proc_init(
	codec_studio_t	*codec)
{
	int	error;
	char	ename[128];

#if BACKTRACE
snd_printk ("codec_studio: backtrace: snd_codec_studio_proc_init\n");
#endif

	sprintf (ename, "studio_codec%u", codec->icodec);
	if (!(codec->entry_proc = snd_info_create_entry (
					codec->card, ename)))
		return -ENOMEM;

	codec->entry_proc->private_data		= codec;
	codec->entry_proc->mode			= S_IFREG | S_IRUGO | S_IWUSR;
	codec->entry_proc->t.text.read_size	= 512;
	codec->entry_proc->t.text.read		= snd_studio_codec_proc_read;

	if ((error=snd_info_register(codec->entry_proc))){
		snd_info_free_entry(codec->entry_proc);
		return error;
	}

	return 0;
}

static void snd_codec_studio_proc_done(
	codec_studio_t	*codec)
{
	if (!codec->entry_proc)
		return;

	snd_info_unregister(codec->entry_proc);
	codec->entry_proc = NULL;
}


/**************************************************************/
#define WAIT_WHILE_DSP_BUSY	0
#define WAIT_WHILE_DSP_READY	1
#define WAIT_WHILE_NOT_HTRQ	2
#define WAIT_WHILE_NOT_HRRQ	3
#define WAIT_WHILE_NOT_TRDY	4
#define WAIT_WHILE_NOT_RUN	5
#define WAIT_WHILE_NOT_IDLE	6
static int dsp56301_wait_while(codec_studio_t * codec,
	unsigned wait_while, unsigned long nretry)

{
	unsigned long retry;
	
	switch (wait_while){
		case WAIT_WHILE_DSP_BUSY:
			for (retry=0; (retry<nretry) && TEST_DSP_BUSY(codec); retry++)
				DELAY(1); /* Breathing for DSP or PCI bus */
			if (codec->stat.nretries_dsp_busy < retry)
				codec->stat.nretries_dsp_busy = retry;
			codec->stat.sum_retries_dsp_busy += retry;
			break;
		case WAIT_WHILE_DSP_READY:
			for (retry=0; (retry<nretry) && !TEST_DSP_BUSY(codec); retry++)
				DELAY(1); /* Breathing for DSP or PCI bus */
			if (codec->stat.nretries_dsp_ready < retry)
				codec->stat.nretries_dsp_ready = retry;
			codec->stat.sum_retries_dsp_ready += retry;
			break;
		case WAIT_WHILE_NOT_HTRQ:
			for (retry=0; (retry<nretry) && !TEST_HTRQ(codec); retry++)
				DELAY(1); /* Breathing for DSP or PCI bus */
			if (codec->stat.nretries_not_htrq < retry)
				codec->stat.nretries_not_htrq = retry;
			codec->stat.sum_retries_not_htrq += retry;
			break;
		case WAIT_WHILE_NOT_HRRQ:
			for (retry=0; (retry<nretry) && !TEST_HRRQ(codec); retry++)
				DELAY(1); /* Breathing for DSP or PCI bus */
			if (codec->stat.nretries_not_hrrq < retry)
				codec->stat.nretries_not_hrrq = retry;
			codec->stat.sum_retries_not_hrrq += retry;
			break;
		case WAIT_WHILE_NOT_TRDY:
			for (retry=0; (retry<nretry) && !TEST_TRDY(codec); retry++)
				DELAY(1); /* Breathing for DSP or PCI bus */
			if (codec->stat.nretries_not_trdy < retry)
				codec->stat.nretries_not_trdy = retry;
			codec->stat.sum_retries_not_trdy += retry;
			break;
		case WAIT_WHILE_NOT_RUN:
			for (retry=0; (retry<nretry) && !TEST_RUN_STATE(codec); retry++)
				DELAY(1); /* Breathing for DSP or PCI bus */
			if (codec->stat.nretries_not_run < retry)
				codec->stat.nretries_not_run = retry;
			codec->stat.sum_retries_not_run += retry;
			break;
		case WAIT_WHILE_NOT_IDLE:
			for (retry=0; (retry<nretry) && TEST_RUN_STATE(codec); retry++)
				DELAY(1); /* Breathing for DSP or PCI bus */
			if (codec->stat.nretries_not_idle < retry)
				codec->stat.nretries_not_idle = retry;
			codec->stat.sum_retries_not_idle += retry;
			break;
		default:
	    		snd_printk("codec_studio: dsp56301_wait_while: invalid value\n");
			return -EINVAL;
	}
	if (nretry <= retry)
		return -EBUSY;

	return 0;
}

/**************************************************************/
static int dsp56301_lcommand(
		codec_studio_t	**lcodec,
		unsigned	ncodec,
		unsigned long	command,
		unsigned long	retry_while_ready,
		unsigned long	wait_uS)
{
	int error, idx;

	for (idx=0; idx<ncodec; idx++)
	if ((error = dsp56301_wait_while(
			lcodec[idx], WAIT_WHILE_DSP_BUSY, STUDIO_RETRY))){
		snd_printk("codec_studio: dsp56301_command: Codec #%d quite busy for 0x%lx command\n",
			lcodec[idx]->icodec, command);
		return -EBUSY;
	}

	for (idx=0; idx<ncodec; idx++)
		COMMAND_DSP(lcodec[idx], command);

	/* Detect DSP at work */
	if (retry_while_ready)
		for (idx=0; idx<ncodec; idx++)
			dsp56301_wait_while(
				lcodec[idx], WAIT_WHILE_DSP_READY, retry_while_ready);

	if (wait_uS/10 != 0){
		snd_delay(wait_uS/10);
		for (idx=0; idx<ncodec; idx++)
			lcodec[idx]->stat.sum_wait_uS += wait_uS;
	}

	return 0;
}

/**************************************************************/
static int dsp56301_command(
		codec_studio_t	*codec,
		unsigned long	command,
		unsigned long	retry_while_ready,
		unsigned long	wait_uS)
{
return (dsp56301_lcommand(
		&codec, 1, command,
		retry_while_ready, wait_uS));
}

/**************************************************************/

static int _dsp56301_write(codec_studio_t *codec,
			void *opaque_data, unsigned nbytes,
                        int altera_flag)
{
	unsigned long	nretry, idx;
	unsigned long	dataout;
	unsigned long	*data     = opaque_data;
	unsigned	nitems_ul = nbytes/sizeof(unsigned long);
	int		error;

        nretry = altera_flag ? ALTERA_RETRY : STUDIO_RETRY;
	for (idx=0; idx<nitems_ul; idx++){
	
		if ((error = dsp56301_wait_while(codec, WAIT_WHILE_NOT_HTRQ, nretry))){
			snd_printk("codec_studio: _dsp56301_write: DSP failed to write before index %ld\n", idx);
			return error;
		}

		/* Implementation of following altera logic depends     */
		/* on "merge_altera" program in TOPDIR/utils directory. */
		/* Final unsigned long masked with 0xFF00 as stated     */
		/* on pag. 17, HISv0.9                                  */
		dataout = data ? data[idx] : 0;
		if (altera_flag){
			if (!codec->up_rev_d)
				dataout = dataout >> 8;
			dataout &= 0x00FFL;
			if (idx == nitems_ul-1)
				dataout |= 0xFF00L;
		}
		DSP_U32_OUT(codec, dataout);
	}

	/* Detect DSP ready again */
	if ((error = dsp56301_wait_while(codec, WAIT_WHILE_DSP_BUSY, nretry))){
		snd_printk("codec_studio: _dsp56301_write: DSP quite busy. Status 0x%lx\n",
			(unsigned long)STATUS_DSP(codec));
		return error;
	}

	return 0;
}

/**************************************************************/
#define TX_CHUNK	1024
static int dsp56301_write(
	codec_studio_t	*codec,
	void		*data,
	int		nbytes)
{
	for (; nbytes > 0; nbytes -= TX_CHUNK)	
	if (data){
		DSP_DATA_OUT(codec, data,
			(nbytes >= TX_CHUNK) ? TX_CHUNK : nbytes);
		data   += TX_CHUNK;
	}
	else
		DSP_ZEROED_OUT(codec,
			(nbytes >= TX_CHUNK) ? TX_CHUNK : nbytes);

	return 0;
}

/**************************************************************/
static int dsp56301_read_u32(
	codec_studio_t	*codec,
	u32		*data,
	int		nu32)
{
	unsigned idx;

	for (idx=0; idx<nu32; idx++)
		data[idx] = DSP_U32_IN(codec);

	return 0;
}

#define RX_CHUNK	1024
static int dsp56301_read(
	codec_studio_t	*codec,
	void		*data,
	int		nbytes)
{
#if 1
	for (; nbytes > 0; nbytes -= RX_CHUNK){	
		DSP_DATA_IN (codec, data,
				(nbytes >= RX_CHUNK) ? RX_CHUNK : nbytes);
		data   += RX_CHUNK;
	}
#else
	dsp56301_read_u32(codec, (u32 *)data, nbytes/sizeof(u32));
#endif
	
	return 0;
}


/**************************************************************/

static int dsp56301_command_write(
	codec_studio_t	*codec,
	unsigned long	command,
	unsigned long	wait_uS,
	void		*opaque_data,
	unsigned	nbytes)
{
	int error;

	if ((error = dsp56301_command(codec, command, 100, wait_uS)))
		return error;
		
	if ((error = dsp56301_write(codec, opaque_data, nbytes)))
		return error;

	return 0;
}

/**************************************************************/

static int dsp56301_command_read(
	codec_studio_t	*codec,
	unsigned long	command,
	unsigned long	wait_uS,
	void		*opaque_data,
	unsigned	nbytes)
{
	int error;

	if ((error = dsp56301_command(codec, command, 100, wait_uS)))
		return error;
		
	if ((error = dsp56301_read(codec, opaque_data, nbytes)))
		return error;

	return 0;
}

/**************************************************************/

static int dsp56301_get_status(
	codec_studio_t	*codec,
	int		clear_error)
{
	int	error, idx;
	u32	meter;

	if ((error = dsp56301_command_read(codec, HC_GET_STATUS_NMI, 0,
			&codec->status, sizeof(codec->status)))){
		snd_printk("codec_studio: dsp56301_get_status: Failed to read\n");
		return error;
	}

	/* Extend sign for 24-bits signed meters */
	for (idx=0; idx<16; idx++){
		meter = codec->status.peak_playback[idx];
		if (meter & 0x00800000){
			meter = ((~meter)&0x00ffffff)+0x01;
			meter = (~meter)+0x01;
			codec->status.peak_playback[idx]=meter;
		}
		meter = codec->status.peak_record[idx];
		if (meter & 0x00800000){
			meter = ((~meter)&0x00ffffff)+0x01;
			meter = (~meter)+0x01;
			codec->status.peak_record[idx]=meter;
		}
		if (idx>1) continue;
		meter = codec->status.peak_monitor[idx];
		if (meter & 0x00800000){
			meter = ((~meter)&0x00ffffff)+0x01;
			meter = (~meter)+0x01;
			codec->status.peak_monitor[idx]=meter;
		}
	}
	
	if (!clear_error)
		return 0;
	
	if (codec->status.iostat & NO_ADAT_SYNC)
		goto __clear_err;
	if (codec->status.iostat & INPUT_A_DATA_ERROR)
		goto __clear_err;
	if (codec->status.iostat & INPUT_B_DATA_ERROR)
		goto __clear_err;
	if (codec->status.error  & GENERIC_ERROR)
		goto __clear_err;
	if (codec->status.error  & DRAM_ERROR)
		goto __clear_err;
		
	return 0;
	
      __clear_err:
	if ((error = dsp56301_command(codec, HC_CLEAR_ERR_NMI, 100, 100)))
		return error;
      
	return error;
}

/**************************************************************/
#define TRANSITION_TO_IDLE	0
#define TRANSITION_TO_RUN	1
static int dsp56301_state_transition (codec_studio_t *codec, int transition)

{
	int needed, error, wait_for;
	unsigned long command, nretry;
	
	needed = (transition == TRANSITION_TO_IDLE)
		? TEST_RUN_STATE(codec) : TEST_IDLE_STATE(codec);
	if (!needed) return 0;

	command  = (transition == TRANSITION_TO_IDLE)
		? HC_IDLE_NMI : HC_RUN;
	nretry   = (transition == TRANSITION_TO_IDLE)
		? ALTERA_RETRY : STUDIO_RETRY;
	wait_for = (transition == TRANSITION_TO_IDLE)
		? WAIT_WHILE_NOT_IDLE : WAIT_WHILE_NOT_RUN;
		
	if ((error = dsp56301_command(codec, command, 0, 0)))
		return error;
	if ((error = dsp56301_wait_while(codec, wait_for, nretry))){
		snd_printk("codec_studio: dsp56301_state_transition: Impossible transition to %s.\n",
			(transition == TRANSITION_TO_IDLE) ? "Idle State" : "Run State");
		return error;
		}
	
	return 0;
}

/**************************************************************/

static int dsp56301_test_and_reset(codec_studio_t * codec)
{
	int error;
	
	/* Ready ?? */
	if (dsp56301_wait_while(codec, WAIT_WHILE_NOT_TRDY, STUDIO_RETRY))
		return -ENODEV;
	if (TEST_TRDY(codec))
		return 0;
	
	if ((error = dsp56301_command(codec, HC_RESET_DSP_NMI, 100, 500)))
		return error;

	if (dsp56301_wait_while(codec, WAIT_WHILE_NOT_TRDY, STUDIO_RETRY))
		return -ENODEV;

	return 0;
}

/**************************************************************/

static int dsp56301_update(
	codec_studio_t	*codec,
	unsigned	nformat)

{
	/*     -1. Transition to Idle State
	 *	0. Stop serial data
	*	1. Get current param group
  	*	2. Get current setup group
  	*	3. Mask in new values for above
  	*	4. Program FPGA with proper program
  	*	5. Set FPGA bytes as needed
  	*	6. Write updated setup group
  	*	7. Start serial data
  	*	8. Write updated param group
	*/

	int		error;
	unsigned long	dataout, format_mask;
	int		DAC_attenuation;
	int		DAC_lowfs;
	int		DAC_resolution18;


	/********/
  	/* -1. Transition to Idle State */
	/********/
	if ((error = dsp56301_state_transition (codec, TRANSITION_TO_IDLE)))
		return error;

	/********/
  	/* 0. Stop serial data */
	/********/
	if ((error = dsp56301_command(codec, HC_STOP_SERIAL, 0, 200)))
		return error;


	/********/
	/* 1. Get current param group */
	/********/
	if ((error = dsp56301_command_read(codec, HC_GET_PARAMS, 0,
			&codec->params, sizeof(codec->params)))){
		snd_printk("codec_studio: dsp56301_update: Failed to read params group\n");
		return error;
	}


	/********/
  	/* 2. Get current setup group */
	/********/
	if ((error = dsp56301_command_read(codec, HC_GET_SETUP, 0,
			&codec->setup, sizeof(codec->setup)))){
		snd_printk("codec_studio: dsp56301_update: Failed to read setup group\n");
		return error;
	}

	/********/
  	/* 3. Mask in new values for above */
	/********/

	/* Setup group control  Pag. 7-9, HISv0.9  */
	switch (nformat){
		case 0: /* 2ADAT */
		case 5: /* 2ADAT96 */
			codec->setup.format	= 0x00;
			codec->setup.nplay	= 16;
			codec->setup.nrecord	= 16;
			break;
		default:
	    		snd_printk("codec_studio: dsp56301_update: invalid format code\n");
			return -EINVAL;
	}

	/* 24-bit audio mode for 2ADAT, pag. 22, HISv0.9 */
	codec->setup.format |= 0x100;

	/* Enable timecode updates for master boards and */
	/* SYNCIO mode                                   */
	if (codec->syncio > 0)
		codec->setup.format |= TCEN;

	/* Now mask in any extra bits as passed in the format mask */
	/* The format mask is currently unused.                    */
	format_mask = 0L;
	codec->setup.format |= format_mask & 0xfffffe00;

	/* Todo: translations linear monitor selections to */
	/* format-specific bitmaps			   */
	codec->setup.mon1l = 0;
	codec->setup.mon2l = 0;
	codec->setup.mon1r = 0;
	codec->setup.mon2r = 0;
/*
	codec->setup.rate_divider = 0;
*/

	/* ToDo, param group control  Pag. 10-11, HISv0.9  */

	memset (codec->params.output_monitor_level, 0, sizeof(codec->params.output_monitor_level));
	memset (codec->params.master_monitor_level, 0, sizeof(codec->params.master_monitor_level));
	memset (codec->params.input_monitor_level,  0, sizeof(codec->params.input_monitor_level));

/* Tmp */
codec->params.master_monitor_level[0] = 0x01<<18;
codec->params.master_monitor_level[1] = 0x01<<18;

	/* Now we have separate settings for right and left channels, mapped into   */
	/* the nDacAttenuation word like this:  The left channel is in the LS byte, */ 
	/* and the right channel is the XOR of the left channel and the 2nd LS byte */
	DAC_attenuation  = 0;
	DAC_lowfs        = 1;
	DAC_resolution18 = 0;
  	codec->params.DA_control  = ((DAC_attenuation & 0xff)^((DAC_attenuation & 0xff00)>>8)) << 16;
	codec->params.DA_control |= (DAC_attenuation & 0xff)<<8;
	codec->params.DA_control |= 0x800000;  /* OR in MS bit to select R channel */
	codec->params.DA_control |= DAC_lowfs        ? 0x00 : 0x80;
	codec->params.DA_control |= DAC_resolution18 ? 0x10 : 0x00;

/* Tmp */
codec->params.DA_control  = 0x800000;

	codec->params.output_phases = 0;
	codec->params.input_phases  = 0;


	/********/
  	/* 4. Program FPGA with proper program */
	/********/

	switch (nformat){
		case 0: /* 2ADAT */
		case 5: /* 2ADAT96 */
			if ((error = dsp56301_command (codec, HC_REPROG_FPGA,
					100, 0)))
				return error;
			if ((error = _dsp56301_write(codec, &altera2020[0],
					sizeof(altera2020), 1)))
				return error;
			break;
		default:
	    		snd_printk("codec_studio: dsp56301_update: invalid format code\n");
			return -EINVAL;
	}


	/********/
  	/* 5. Set FPGA bytes as needed */
	/********/

	/* ToDo, support for other clock sources */
	switch (codec->adat_sync){
		case -1:
		case  0:
			dataout = CLK_PLL_OPTICAL_A;
			break;
		case  1:
			dataout = CLK_PLL_OPTICAL_B;
			break;
		case  2:
			dataout = CLK_PLL_EXTERNAL;
			break;
		case  3:
			dataout = CLK_44K;
			break;
		case  4:
			dataout = CLK_48K;
			break;
		case  5:
			dataout = CLK_PLL_TIMER1;
			break;
		default:
	    		snd_printk("codec_studio: dsp56301_update: invalid adat sync\n");
			return -EINVAL;
	}
	if ((error = dsp56301_command_write(codec, HC_SEND_FPGABYTE, 0, &dataout, sizeof(dataout)))){
		snd_printk("codec_studio: dsp56301_update: Failed to write FPGA byte\n");
		return error;
	}


	/********/
  	/* 6. Write updated setup group */
	/********/
	if ((error = dsp56301_command_write(codec, HC_SEND_SETUP, 0, &codec->setup, sizeof(codec->setup)))){
		snd_printk("codec_studio: dsp56301_update: Failed to write setup group\n");
		return error;
	}

	/********/
	/* 7. Start serial data */
	/********/
	if ((error = dsp56301_command(codec, HC_START_SERIAL, 100, 1000)))
		return error;

	
	/********/
  	/* 8. Write updated params group */
	/********/
	if ((error = dsp56301_command_write(codec, HC_SEND_PARAMS, 0, &codec->params, sizeof(codec->params)))){
		snd_printk("codec_studio: dsp56301_update: Failed to write params group\n");
		return error;
	}

	return 0;
}

/**************************************************************/

static int dsp56301_boot(codec_studio_t * codec)
{
	unsigned long delay, retry, dummy;
	int success, error;

	CONTROL_DSP(codec, HTF01|HRF01);	/* 24lsbs to/from PCI bus */
	COMMAND_DSP(codec, 0L); 		/* Innocuous value */

	/* Cleannig DSP. Buggy ??? */
	for (retry=0; (retry<STUDIO_RETRY) && TEST_HRRQ(codec); retry++)
		dummy = DSP_U32_IN(codec);
	if (TEST_HRRQ(codec) || !TEST_TRDY(codec) || TEST_DSP_BUSY(codec)){
		COMMAND_DSP(codec, HC_RESET_DSP_NMI);
	    	snd_delay(500);
		}

	/* Wait until the HI is completely empty */
	if ((error = dsp56301_test_and_reset(codec))) {
		snd_printk("codec_studio: dsp56301_boot: DSP failed to reset\n");
		return error;
	}

	/* Try to write first word of boot code */
        for  (success=retry=0; !success && (retry<10); retry++){
		DSP_U32_OUT(codec, bootcode[0]);
		for (delay=0; (delay<TINY_DELAY) && TEST_TRDY(codec); delay++);
		success = (dsp56301_test_and_reset(codec) == 0);
	}
	if (!success) {
		snd_printk("codec_studio: dsp56301_boot: DSP failed to boot\n");
		return -ENODEV;
	}

	/* Write the rest of the boot code */
	if ((error = dsp56301_write(codec, &bootcode[1], sizeof(bootcode)- sizeof(bootcode[0])))){
		snd_printk("codec_studio: dsp56301_boot: DSP failed to write boot code\n");
		return error;
	}

	/* Wait for the cookie. This timeout is 1024 times longer standard  */
	/* because the dsp must finish lengthy power-on testing of the DRAM */
	/* Actually, cookie data is unused                                  */
	if ((error = dsp56301_wait_while(codec, WAIT_WHILE_NOT_HRRQ, STUDIO_RETRY<<10))){
		snd_printk("codec_studio: dsp56301_boot: DSP not ready to read\n");
		return error;
	}
	if ((error = dsp56301_read(codec, &codec->cookie[0], sizeof(codec->cookie)))){
		snd_printk("codec_studio: dsp56301_boot: DSP failed to read cookie\n");
		return error;
	}
	codec->cookie[0] |= codec->cookie[1] << 24;
	codec->cookie[1]  = codec->cookie[1] >>  8;


	/* Get status after booting */
	if ((error = dsp56301_get_status(codec, 1))){
		snd_printk("codec_studio: dsp56301_boot: DSP failed to manage status group\n");
		return error;
	}
	codec->up_rev_d = (codec->status.error & UP_REV_D);

	
	/* Update programming stuff for DSP */
	/* ToDo : support for 2ADAT96 */
	if ((error = dsp56301_update(codec, MODE_2ADAT)))
		return error;

	return 0;
}

/**************************************************************/

static int dsp56301_shutdown(codec_studio_t * codec)
{
	int error;
	
	if ((error = dsp56301_state_transition (codec, TRANSITION_TO_IDLE)))
		return error;
		
	if (dsp56301_command(codec, HC_RESET_DSP_NMI, 100, 500))
		return -EBUSY;

	return 0;
}

/**************************************************************/
unsigned int snd_codec_get_channel_size(void)

{
	return (CHANNEL_NBYTES);
}


/**************************************************************/

/* buffer must be 16*CHANNEL_NBYTES-sized or NULL */
int snd_codec_studio_playback(
		codec_studio_t *codec,
		void *buffer)
{
	return (dsp56301_command_write(codec, HC_WRITE_SAMPLES_NMI, 0,
			buffer, 16*CHANNEL_NBYTES));
}

/**************************************************************/

/* buffer must be 16*CHANNEL_NBYTES-sized or NULL */
int snd_codec_studio_record(
		codec_studio_t *codec,
		void *buffer)
{
	int error;
	
	if (!buffer)
		return 0;
#if 0
	return (error = dsp56301_command_read(
				codec, HC_READ_SAMPLES_NMI, 0,
				buffer, 16*CHANNEL_NBYTES));
#else
	if (codec->bitmap_read & 0x000f)
	if ((error = dsp56301_command_read(
			codec, HC_READ_BLOCK_0_NMI, 0,
			buffer, 4*CHANNEL_NBYTES)))
		return error;
	buffer += 4*CHANNEL_NBYTES;

	if (codec->bitmap_read & 0x00f0)
	if ((error = dsp56301_command_read(
			codec, HC_READ_BLOCK_1_NMI, 0,
			buffer, 4*CHANNEL_NBYTES)))
		return error;
	buffer += 4*CHANNEL_NBYTES;

	if (codec->bitmap_read & 0x0f00)
	if ((error = dsp56301_command_read(
			codec, HC_READ_BLOCK_2_NMI, 0,
			buffer, 4*CHANNEL_NBYTES)))
		return error;
	buffer += 4*CHANNEL_NBYTES;

	if (codec->bitmap_read & 0xf000)
	if ((error = dsp56301_command_read(
			codec, HC_READ_BLOCK_3_NMI, 0,
			buffer, 4*CHANNEL_NBYTES)))
		return error;
#endif

	return 0;
}

/**************************************************************/

/* buffer must be 2*CHANNEL_NBYTES-sized or NULL */
int snd_codec_studio_monitor(
		codec_studio_t *codec,
		void *buffer)
{
	return (dsp56301_command_write(codec, HC_WR_MONITOR_NMI, 0,
			buffer, 2*CHANNEL_NBYTES));
}

/**************************************************************/

int snd_codec_studio_status(
		codec_studio_t *codec)
{
	int error;

	if ((error = dsp56301_get_status(codec, 1))){
		snd_printk("codec_studio: snd_codec_studio_isr: Failed to get status\n");
		return error;
	}

	return 0;
}

/**************************************************************/

int snd_codec_studio_clear_error(
	codec_studio_t	**lcodec,
	unsigned	ncodec)
{
	int		error, idx;
	codec_studio_t	*codec;

	for (idx=0; idx<ncodec; idx++){
		codec = lcodec[idx];

		CLEAR_RT_ERROR(codec);

		if ((error = dsp56301_command_write(codec,
				HC_SEND_STATUS_NMI, 0,
				&codec->status,
				sizeof(codec->status))))
			return error;
	}

	return 0;
}

/**************************************************************/

int snd_codec_studio_test_int(
		codec_studio_t *codec)
{
	return (TEST_HINT(codec));
}

/**************************************************************/

int snd_codec_studio_ackint(
		codec_studio_t *codec)
{
	return (dsp56301_command(codec, HC_ACK_HOSTINT_NMI, 0, 0));
}

/**************************************************************/

int snd_codec_studio_start(
	codec_studio_t	**lcodec,
	unsigned	ncodec)
{
	int error, idx;

	for (idx=0; idx<ncodec; idx++){
		if ((error = dsp56301_state_transition (
				lcodec[idx], TRANSITION_TO_RUN)))
			return error;
		lcodec[idx]->running = 1;
	}
			
	return 0;
}

/**************************************************************/

int snd_codec_studio_stop(
	codec_studio_t	**lcodec,
	unsigned	ncodec)
{
	int error, idx;
	
	for (idx=0; idx<ncodec; idx++){
		if ((error = dsp56301_state_transition (
				lcodec[idx], TRANSITION_TO_IDLE)))
			return error;
		lcodec[idx]->running = 0;
	}
		
	return 0;
}

/**************************************************************/
int snd_codec_studio_sync_serial(
	codec_studio_t	**lcodec,
	unsigned	ncodec)
{
	int error;

	if ((error=snd_codec_studio_stop(lcodec, ncodec)))
		return error;
	if ((error = dsp56301_lcommand(lcodec, ncodec,
			HC_STOP_SERIAL, 0, 200)))
		return error;
	if ((error = dsp56301_lcommand(lcodec, ncodec,
			HC_START_SERIAL, 100, 1000)))
		return error;

	return 0;
}

/***************************************************************/
int snd_codec_studio_test_routes(
	codec_studio_t * codec)

{
	unsigned idx;

	for (idx=0; idx<16; idx++){
		if (codec->playback_routing[idx])
			return 1;
		if (codec->record_routing[idx])
			return 1;
		if (idx > 1) continue;
		if (codec->monitor_routing[idx])
			return 1;
	}

	return 0;
}

/**************************************************************/

void snd_codec_studio_free(codec_studio_t * codec)
{
	if (!codec)
		return;

	snd_codec_studio_proc_done(codec);

	if (codec->ba_reg)
		dsp56301_shutdown(codec);
	if (codec->ba_reg)
		iounmap(codec->ba_reg);
	if (codec->ba_data)
		iounmap(codec->ba_data);

	snd_kfree(codec);
}

/**************************************************************/

codec_studio_t *snd_codec_studio_create (
	snd_card_t	*card,
	unsigned	icodec,
	unsigned long	pci_base_address)
{
	codec_studio_t *codec;

	if (!(codec = (codec_studio_t *) snd_kcalloc(sizeof(*codec), GFP_KERNEL)))
		return NULL;
	codec->card		= card;
	codec->icodec		= icodec;
	codec->adat_sync	= adat_sync[icodec];
	codec->syncio		= syncio[icodec];
	codec->running		= 0;
	codec->access_lock	= SPIN_LOCK_UNLOCKED;

	/* Map PCI memory address space */
	codec->ba_reg  = ioremap(pci_base_address + DSP56301_REG_OFFSET,  DSP56301_REG_SIZE);
	codec->ba_data = ioremap(pci_base_address + DSP56301_DATA_OFFSET, DSP56301_DATA_SIZE);
	if (codec->ba_reg == NULL || codec->ba_data == NULL) {
	        snd_codec_studio_free(codec);
	    	return NULL;
	}


	/* Boot DSP56301 */
	if (dsp56301_boot(codec)) {
	        snd_codec_studio_free(codec);
	    	return NULL;
	}

	if (snd_codec_studio_proc_init(codec)){
		snd_codec_studio_free (codec);
		return NULL;
	}

	return codec;
}
