#include "studio.h"
#include "codec_studio.h"
#include "pcm_studio.h"
#include <linux/vmalloc.h>

extern int monitor_on;

static void handler_timer (unsigned long cd);


/***************************************************************/
static unsigned long time_diff (
	struct timeval	*reference,
	struct timeval	*now)
{
	do_gettimeofday(now);
	return ((reference == NULL) ? 0:
			(now->tv_sec  - reference->tv_sec)*1000000
		       +(now->tv_usec - reference->tv_usec));
}


/***************************************************************/
/* snd_studio_sample_rate() function updates the "real_rate"   */
/* field of a subchannel using the last interrupt period       */
/* measurement. Also, return the divisor integer to get the    */
/* nearest rate to the desired rate (subchannel1->rate). On    */
/* error or exception, this function is extremely prophylactic,*/
/* returning non-problematic values.                           */
/* Run-time changes of sample are supported.                   */
/***************************************************************/
static int snd_studio_sample_rate(
		snd_pcm_studio_t	*pcm_s,
		snd_pcm_subchn_t	*subchn)
{
	snd_pcm1_subchn_t	*subchn1;
	unsigned long		real_rate;
	unsigned long		divisor, div10;

	if (!(subchn1 = subchn->private_data))
		return 1;

	real_rate = (pcm_s->stat.period_int == 0)
			? 0 : (250000*pcm_s->frame_per_channel)/pcm_s->stat.period_int;
	if ((real_rate == 0) || (real_rate <= subchn1->rate)){
		subchn1->real_rate = subchn1->rate;
		return 1;
	}
	div10   = (10*real_rate)/subchn1->rate;
	divisor = real_rate/subchn1->rate;
	if (div10%10 > 5) divisor++;
	
	subchn1->real_rate = real_rate/divisor;
	return divisor;
}

/*********************************************************/
/* One "pcm_studio" subsystem manages one or more codec. */
/* If there is a master card, the master pcm subsystem   */
/* manages all of codecs. Else every pcm manages its     */
/* codec.                                                */
/*                                                       */
/* Block size of an Alsa subchannel is forced to contain */
/* whole data for a channel of a codec. An interrupt     */
/* reads and writes every channel of every codec, and    */
/* transfer to/from opened PCM-Alsa subchannels data     */
/* blocks with fixed size.                               */
/*********************************************************/
static void snd_studio_pcm_proc_read(
	snd_info_buffer_t	*buffer,
	void			*private_data)
{
	snd_pcm_studio_t	*pcm_s;
	snd_pcm_subchn_t	*subchn;
	snd_pcm1_subchn_t	*subchn1;

	pcm_s = (snd_pcm_studio_t *) private_data;

	snd_iprintf(buffer, "Sonorus Studi/o Statistics\n");
	snd_iprintf(buffer, "Interrupts\t\t\t: %lu\n", pcm_s->stat.nint);
	snd_iprintf(buffer, "Real-Time Errors\t\t: %lu\n", pcm_s->stat.nrt_errors);
	snd_iprintf(buffer, "Mismatched Codecs Errors\t: %lu\n", pcm_s->stat.mismatched_codecs);
	snd_iprintf(buffer, "Real Interrupt Period (uS)\t: %lu (%lu Hz)\n",
				pcm_s->stat.period_int,
				(pcm_s->stat.period_int == 0)
					? 0 : (250000*pcm_s->frame_per_channel)/pcm_s->stat.period_int);
	snd_iprintf(buffer, "Time of Service (uS)\t\t: %lu (%lu%)\n",
				pcm_s->stat.time_service,
				(pcm_s->stat.period_int == 0)
					? 0 : (100*pcm_s->stat.time_service)/pcm_s->stat.period_int);
	snd_iprintf(buffer, "Time Unsync Codecs (S)\t\t: %lu\n", pcm_s->stat.time_unsync);
	snd_iprintf(buffer, "Debug point\t\t\t: %lu\n", pcm_s->stat.debug_point);

	snd_iprintf(buffer, "Opened playback channels\t:");
	for (subchn=pcm_s->pcm->playback.subchn; subchn; subchn=subchn->next)
		if ((subchn1=subchn->private_data))
			snd_iprintf(buffer, " %d(%u)",
				subchn->number, subchn1->real_rate);
	snd_iprintf(buffer, "\n");
	snd_iprintf(buffer, "Opened record channels\t\t:");
	for (subchn=pcm_s->pcm->capture.subchn; subchn; subchn=subchn->next)
		if ((subchn1=subchn->private_data))
			snd_iprintf(buffer, " %d(%u)",
				subchn->number, subchn1->real_rate);
	snd_iprintf(buffer, "\n");
}

/*********************************************************/
static int snd_pcm_studio_proc_init(
	snd_pcm_studio_t	*pcm_s)
{
	int	error;
	char	ename[128];

	sprintf (ename, "studio_pcm%u", pcm_s->ipcm);
	if (!(pcm_s->entry_proc = snd_info_create_entry (
					pcm_s->card, ename)))
		return -ENOMEM;

	pcm_s->entry_proc->private_data		= pcm_s;
	pcm_s->entry_proc->mode			= S_IFREG | S_IRUGO | S_IWUSR;
	pcm_s->entry_proc->t.text.read_size	= 512;
	pcm_s->entry_proc->t.text.read		= snd_studio_pcm_proc_read;
	if ((error=snd_info_register(pcm_s->entry_proc))){
		snd_info_free_entry(pcm_s->entry_proc);
		return error;
	}

	return 0;
}

/*********************************************************/
static void snd_pcm_studio_proc_done(
	snd_pcm_studio_t	*pcm_s)
{
	if (!pcm_s->entry_proc)
		return;

	snd_info_unregister(pcm_s->entry_proc);
	pcm_s->entry_proc = NULL;
}

/*********************************************************/
static int handler_codec_status (
	snd_pcm_studio_t *pcm_s)

{
	int	icodec, error_sync;

	/* Update status */
	error_sync = pcm_s->on_rt_error = 0;
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++){
		if (snd_codec_studio_status(pcm_s->codec[icodec]))
			goto __critical_bad_status;
		error_sync |= TEST_ADAT_SYNC_ERROR(pcm_s->codec[icodec]);
		pcm_s->on_rt_error |= TEST_RT_ERROR(pcm_s->codec[icodec]);
	}

	/* If any codec without adat sync, stop codecs and    */
	/* (re)start timer. Real-time errors are not critical */
	/* bad status.                                        */
	if (error_sync){
		pcm_s->stat.time_unsync++;
		goto __critical_bad_status;
	}

	return 0;

      __critical_bad_status:
	snd_codec_studio_stop(pcm_s->codec, pcm_s->ncodecs);

	init_timer(&pcm_s->t_adat_sync);
	pcm_s->t_adat_sync.function	= handler_timer;
	pcm_s->t_adat_sync.data		= (unsigned long) pcm_s;
	pcm_s->t_adat_sync.expires	= jiffies + HZ; /* 1 sec */
	add_timer(&pcm_s->t_adat_sync);
	return 1;
}

/*********************************************************/
static void handler_codec_open (
	snd_pcm_studio_t *pcm_s)
{
	unsigned long flags;

	/* If bad status, no start codecs and return */
	if (handler_codec_status (pcm_s))
		return;

	/* There should not be active irqs. Anyway, */
	/* prophylaxis is always good.              */
	spin_lock_irqsave(&pcm_s->access_lock, flags);

	/* Sync and Start codecs */
	snd_codec_studio_sync_serial(pcm_s->codec, pcm_s->ncodecs);
	snd_codec_studio_start (pcm_s->codec, pcm_s->ncodecs);

	spin_unlock_irqrestore(&pcm_s->access_lock, flags);
}

/*********************************************************/
static void handler_codec_close (
	snd_pcm_studio_t *pcm_s)
{
	if (timer_pending(&pcm_s->t_adat_sync))
		del_timer(&pcm_s->t_adat_sync);
	snd_codec_studio_stop(
		pcm_s->codec, pcm_s->ncodecs);
}

/*********************************************************/
static void handler_timer (unsigned long cd)
{
	snd_pcm_studio_t *pcm_s = (snd_pcm_studio_t *) cd;

	handler_codec_open(pcm_s);
}

/*********************************************************/
#define NBLOCKS	(32*MAX_NSAMPLES/NSAMPLES)
static int snd_studio_open(
		void			*private_data,
		snd_pcm_subchn_t	*subchn)
{
	snd_pcm_studio_t	*pcm_s;
	snd_pcm1_subchn_t	*subchn1;
	
	if (!(pcm_s = private_data))
		return -EINVAL;
	if (!pcm_s->ncodecs)
		return -ENODEV;
	if (pcm_s->error_codec)
		return -ENODEV;
	subchn1 = subchn->private_data;

	subchn1->dmaptr		= NULL;
	subchn1->frag_size	= 0;
	if (!(subchn1->private_data = snd_kcalloc(
					sizeof(snd_studio_data_subchn1_t),
					GFP_KERNEL)))
		return -ENOMEM;

	subchn1->size = NBLOCKS*pcm_s->frame_per_channel;
	if (!(subchn1->buffer = vmalloc(subchn1->size)))
		return -ENOMEM;

	snd_pcm_refresh_bitmaps_read (pcm_s);

	if (pcm_s->usage == 0)
		handler_codec_open(pcm_s);
	pcm_s->usage++;

	return 0;
}

/***************************************************************/
static void _snd_studio_close(
		snd_pcm_studio_t	*pcm_s,
		snd_pcm1_subchn_t	*subchn1)
{
	pcm_s->usage--;
	if (pcm_s->usage == 0)
		handler_codec_close(pcm_s);

	if (subchn1->buffer)
		vfree(subchn1->buffer);
	subchn1->buffer = NULL;

	if (subchn1->private_data)
		snd_kfree(subchn1->private_data);
	subchn1->private_data = NULL;
}

static void snd_studio_close(
		void			*private_data,
		snd_pcm_subchn_t	*subchn)
{
	snd_pcm_studio_t	*pcm_s;
	unsigned long		flags;

	if (!(pcm_s = private_data))
		return;

	spin_lock_irqsave(&pcm_s->access_lock, flags);
	_snd_studio_close(pcm_s, subchn->private_data);
	snd_pcm_refresh_bitmaps_read (pcm_s);
	spin_unlock_irqrestore(&pcm_s->access_lock, flags);
}


/***************************************************************/

static int snd_studio_ioctl(
	void			*private_data,
	snd_pcm_subchn_t	*subchn,
	unsigned int		cmd,
	unsigned long		*arg)
{
	snd_pcm_studio_t	*pcm_s;
	snd_pcm1_subchn_t	*subchn1;
	snd_studio_data_subchn1_t *data;

	if (!(pcm_s = private_data))
		return -EINVAL;
	subchn1 = subchn->private_data;

	switch (cmd) {
	case SND_PCM1_IOCTL_RATE:
		snd_studio_sample_rate(pcm_s, subchn);
		data = subchn1->private_data;
		data->idx_sample = 0; /* Reset index */
		return 0;
	case SND_PCM1_IOCTL_FRAG: /* fixed size, lot of code depends on it  */
		*(unsigned int *)arg  = pcm_s->frame_per_channel;
		return 0;
	}
	return -ENXIO;
}


/***************************************************************/
static void snd_studio_prepare(
	void			*private_data,
	snd_pcm_subchn_t	*subchn,
	unsigned char		*buffer,
	unsigned int		size,
	unsigned int		offset,
	unsigned int		count)
{
}


/***************************************************************/
static void snd_studio_trigger(
		void			*private_data,	
		snd_pcm_subchn_t	*subchn,
		int			up)
{
	snd_pcm_studio_t	*pcm_s;
	snd_pcm1_subchn_t	*subchn1;
	unsigned long		flags;
	snd_studio_data_subchn1_t *data;

	if (!(pcm_s = private_data))
		return;
	subchn1 = subchn->private_data;
	data    = subchn1->private_data;

	if (data->open_tag)
		return;

	spin_lock_irqsave(&pcm_s->access_lock, flags);
	data->open_tag = 1;
	spin_unlock_irqrestore(&pcm_s->access_lock, flags);
}

/***************************************************************/
static unsigned int snd_studio_pointer(
			void			*private_data,
			snd_pcm_subchn_t	*subchn,
			unsigned int		used_size)
{
	return used_size;
}


/***************************************************************/

static struct snd_stru_pcm1_hardware snd_studio_playback =
{
	0,			/* flags */
	SND_PCM_FMT_U32_LE,	/* formats */
	SND_PCM_FMT_U32_LE,	/* hardware formats */
	4,			/* align value */
	8,			/* minimal fragment */
	4000,			/* min. rate */
	48000,			/* max. rate */
	1,			/* max. voices */
	snd_studio_open,
	snd_studio_close,
	snd_studio_ioctl,
	snd_studio_prepare,
	snd_studio_trigger,
	snd_studio_pointer,
	snd_pcm1_playback_dma,
	snd_pcm1_dma_move,
	snd_pcm1_playback_dma_neutral
};

static struct snd_stru_pcm1_hardware snd_studio_record =
{
	0,			/* flags */
	SND_PCM_FMT_U32_LE,	/* formats */
	SND_PCM_FMT_U32_LE,	/* hardware formats */
	4,			/* align value */
	8,			/* minimal fragment */
	4000,			/* min. rate */
	48000,			/* max. rate */
	1,			/* max. voices */
	snd_studio_open,
	snd_studio_close,
	snd_studio_ioctl,
	snd_studio_prepare,
	snd_studio_trigger,
	snd_studio_pointer,
	snd_pcm1_capture_dma,
	snd_pcm1_dma_move,
	NULL
};

/***************************************************************/
/* This routine executes the routing logic described in file   */
/* "st_switch.c".                                              */
/***************************************************************/
static void snd_pcm_studio_routing(
	snd_pcm_studio_t	*pcm_s)
{
	unsigned	icodec, itbl, isource;
	codec_studio_t	*codec;
	void		*dest_buf, *source_buf;

	void		*bufplayback = pcm_s->bufplayback;
	void		*bufrecord   = pcm_s->bufrecord;
	void		*bufmonitor  = pcm_s->bufmonitor;


	/* Routing to Playback */
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((codec=pcm_s->codec[icodec])){
		for (itbl=0; itbl<16; itbl++)
		if ((isource=codec->playback_routing[itbl])){
			dest_buf = bufplayback
				+ icodec*pcm_s->frame_per_codec
				+ itbl*pcm_s->frame_per_channel;
			source_buf = (isource <= 16*pcm_s->ncodecs)
				? bufplayback : bufrecord;
			source_buf += ((isource-1) % 16*pcm_s->ncodecs)*pcm_s->frame_per_channel;
			memcpy (dest_buf, source_buf, pcm_s->frame_per_channel);
		}
	}

	/* Routing to Record */
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((codec=pcm_s->codec[icodec])){
		for (itbl=0; itbl<16; itbl++)
		if ((isource=codec->record_routing[itbl])){
			dest_buf = bufrecord
				+ icodec*pcm_s->frame_per_codec
				+ itbl*pcm_s->frame_per_channel;
			source_buf = (isource <= 16*pcm_s->ncodecs)
				? bufplayback : bufrecord;
			source_buf += ((isource-1) % 16*pcm_s->ncodecs)*pcm_s->frame_per_channel;
			memcpy (dest_buf, source_buf, pcm_s->frame_per_channel);
		}
	}

	/* Routing to Monitor */
	if (monitor_on)
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((codec=pcm_s->codec[icodec])){
		for (itbl=0; itbl<2; itbl++)
		if ((isource=codec->monitor_routing[itbl])){
			dest_buf = bufmonitor
				+ icodec*pcm_s->frame_per_codec
				+ itbl*pcm_s->frame_per_channel;
			source_buf = (isource <= 16*pcm_s->ncodecs)
				? bufplayback : bufrecord;
			source_buf += ((isource-1) % 16*pcm_s->ncodecs)*pcm_s->frame_per_channel;
			memcpy (dest_buf, source_buf, pcm_s->frame_per_channel);
		}
	}
}

/***************************************************************/
/* This function loads data from Alsa playback subchannels to  */
/* the playback buffer about writing to codec. When sample     */
/* rate of the subchannel is the maximum (divisor == 1) the    */
/* copy is a simple "memcpy()". Else, each sample from         */
/* subchannel buffer becomes several samples into the playback */
/* buffer (one sample produces "divisor" samples), simulating  */
/* a lower sample rates. Therefore, each block of the          */
/* subchannel buffer expands to data bigger "divisor" times.   */
/* An index of such a data pool is maintained across calls     */
/* using the "private_data" field of the subchannel.           */
/***************************************************************/
static void snd_pcm_load_playback(
	snd_pcm_studio_t *pcm_s)
{
	snd_pcm_subchn_t	*subchn;
	snd_pcm1_subchn_t	*subchn1;
	unsigned long		idx, isrc, idiv, divisor, n_u32;
	void			*bufsrc,  *bufdest;
	u32			*ubufsrc, *ubufdest;
	snd_studio_data_subchn1_t *data;
	
	for (subchn=pcm_s->pcm->playback.subchn; subchn; subchn=subchn->next){
		if (!(subchn1=subchn->private_data))
			continue; /* Not opened */
		if (!(data = subchn1->private_data))
			continue;

		/* If subchannel has just opened-closed and */
		/* Alsa open-close stuff is not completed.  */
		if (!data->open_tag) /* Open tag */
			continue;
		if (!subchn1->buffer) /* Close tag */
			continue;

		bufdest = pcm_s->bufplayback
				+ subchn->number*pcm_s->frame_per_channel;
		bufsrc  = subchn1->buffer
				+ subchn1->tail*subchn1->block_size;
		/* Actual divisor */		
		divisor = snd_studio_sample_rate(pcm_s, subchn);

		/* Case of maximum sample rate */
		if (divisor == 1){
			memcpy(bufdest, bufsrc, pcm_s->frame_per_channel);

			data->idx_sample = 0;
			if (subchn1->ack)
				subchn1->ack(subchn);
			continue;
		}

		/* Case of sample rate other than maximum (divisor != 1) */
		ubufsrc  = (u32 *) bufsrc;
		ubufdest = (u32 *) bufdest;
		isrc  = data->idx_sample;
		n_u32 = pcm_s->frame_per_channel/sizeof(u32);
		for (idx=idiv=0; idx<n_u32; idx++, isrc++){
			/* Unexpected overflow. May be, divisor	*/
			/* was modified.			*/
			idiv = isrc/divisor;
			if (idiv >= n_u32)
				break;
			ubufdest[idx] = ubufsrc[idiv];
		} /* for idx */
		
		if (idiv >= n_u32){
			isrc = 0;
			if (subchn1->ack)
				subchn1->ack(subchn);
		}
		data->idx_sample = isrc;

	} /* for subchn */

}

/***************************************************************/
/* This function behaves analog to "snd_pcm_load_playback ()"  */
/***************************************************************/
static void snd_pcm_unload_record(
	snd_pcm_studio_t *pcm_s)
{
	snd_pcm_subchn_t	*subchn;
	snd_pcm1_subchn_t	*subchn1;
	unsigned long		idx, isrc, idiv, divisor, n_u32;
	void			*bufsrc,  *bufdest;
	u32			*ubufsrc, *ubufdest;
	snd_studio_data_subchn1_t *data;

	for (subchn=pcm_s->pcm->capture.subchn; subchn; subchn=subchn->next){
		if (!(subchn1=subchn->private_data))
			continue; /* Not opened */
		if (!(data = subchn1->private_data))
			continue;

		/* If subchannel has just opened-closed and */
		/* Alsa open-close stuff is not completed.  */
		if (!data->open_tag) /* Open tag */
			continue;
		if (!subchn1->buffer) /* Close tag */
			continue;

		bufsrc  = pcm_s->bufrecord
				+ subchn->number*pcm_s->frame_per_channel;
		bufdest = subchn1->buffer
				+ subchn1->head*subchn1->block_size;
		/* Actual divisor */		
		divisor = snd_studio_sample_rate(pcm_s, subchn);

		/* Case of maximum sample rate */
		if (divisor == 1){
			memcpy(bufdest, bufsrc, pcm_s->frame_per_channel);

			data->idx_sample = 0;
			if (subchn1->ack)
				subchn1->ack(subchn);
			continue;
		}

		/* Case of sample rate other than maximum (divisor != 1) */
		ubufsrc  = (u32 *) bufsrc;
		ubufdest = (u32 *) bufdest;
		isrc  = data->idx_sample;
		n_u32 = pcm_s->frame_per_channel/sizeof(u32);
		for (idx=idiv=0; idx<n_u32; idx++, isrc++){
			if (isrc%divisor)
				continue;
			/* Unexpected overflow. May be, divisor	*/
			/* was modified.			*/
			idiv = isrc/divisor;
			if (idiv >= n_u32)
				break;
			ubufdest[idiv] = ubufsrc[isrc%n_u32];
		} /* for idx */
		
		if (idiv >= n_u32){
			isrc = 0;
			if (subchn1->ack)
				subchn1->ack(subchn);
		}
		data->idx_sample = isrc;

	} /* for subchn */

}

/***************************************************************/
int snd_pcm_studio_interrupt(
	snd_pcm_studio_t *pcm_s)

{
	static struct timeval tv_begin0, tv_begin1, tv_end;

	unsigned	idx, icodec, nint;
	void		*pbuf, *rbuf, *mbuf;

	/* Please, do not insist */
	if (pcm_s->error_codec)
		return 1;

	/* Test for foreign interrupts or wait for all */
	/* of interrupted codecs                       */
	for (idx=0; idx<30; idx++){
		for (nint=icodec=0; icodec<pcm_s->ncodecs; icodec++)
			if (snd_codec_studio_test_int(pcm_s->codec[icodec]))
				nint++;
		if (nint == 0)
			return 1;	/* Foreign interrupt */
		if (nint == pcm_s->ncodecs)
			break;		/* OK, all interrupted */
		snd_delay(1);
	}

	/* Get timing data */
	pcm_s->stat.period_int =
		time_diff ((pcm_s->stat.nint == 0) ? NULL : &tv_begin0,
				&tv_begin1);
	pcm_s->stat.nint++;
	memcpy(&tv_begin0, &tv_begin1, sizeof(tv_begin1));

	/* Not all codecs in interrupt state. In this situation ignore */
	/* real time error indicator.                                  */
	if (nint != pcm_s->ncodecs){
		pcm_s->stat.mismatched_codecs++;
		pcm_s->on_rt_error = 0;
		goto __bye;
	}

	/* On Real-Time error we clear and do a light service */
	if (pcm_s->on_rt_error){
		if (snd_codec_studio_clear_error (pcm_s->codec,
				pcm_s->ncodecs)){
			snd_printk ("pcm_studio: snd_pcm_studio_interrupt: Problems clearing errors\n");
			goto __problems;
		}
		goto __bye;
	}

	/* Reset output buffers */
	memset (pcm_s->bufplayback, 0,
			pcm_s->ncodecs*pcm_s->frame_per_codec);
	if ((monitor_on) || (pcm_s->stat.nint == 1))
		memset (pcm_s->bufmonitor,  0,
				pcm_s->ncodecs*2*pcm_s->frame_per_channel);

	pcm_s->stat.debug_point = pcm_s->codec[0]->bitmap_read;


#if 1
	/* Load data from Alsa playback channels */
	snd_pcm_load_playback(pcm_s);
#endif

#if 1
	/* Record stage */
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++){
		rbuf = pcm_s->bufrecord + icodec*pcm_s->frame_per_codec;
		if (snd_codec_studio_record(pcm_s->codec[icodec], rbuf)){
			snd_printk ("pcm_studio: snd_pcm_studio_interrupt: Problems at record stage\n");
			goto __problems;
		}
	}
#endif

#if 1
	/* Routing stage */
	snd_pcm_studio_routing(pcm_s);
#endif

#if 1
	/* Playback stage */
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++){
		pbuf = pcm_s->bufplayback + icodec*pcm_s->frame_per_codec;
		if (snd_codec_studio_playback(pcm_s->codec[icodec], pbuf)){
			snd_printk ("pcm_studio: snd_pcm_studio_interrupt: Problems at playback stage\n");
			goto __problems;
		}
	}
#endif

#if 1
	/* Monitor stage */
	if ((monitor_on) || (pcm_s->stat.nint == 1))
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++){
		mbuf = pcm_s->bufmonitor + icodec*2*pcm_s->frame_per_channel;
		if (snd_codec_studio_monitor(pcm_s->codec[icodec], mbuf)){
			snd_printk ("pcm_studio: snd_pcm_studio_interrupt: Problems at monitor stage\n");
			goto __problems;
		}
	}
#endif

#if 1
	/* Unload data to Alsa record channels */
	snd_pcm_unload_record (pcm_s);
#endif


      __bye:
	/* Acknowledge stage to codecs */
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++){
		if (snd_codec_studio_ackint  (pcm_s->codec[icodec])){
			snd_printk ("pcm_studio: snd_pcm_studio_interrupt: Problems at acknowledge stage\n");
			goto __problems;
		}
	}


	/* Status stage (each 1 sec, roughly) */
#if 1
	if ((pcm_s->on_rt_error)
		|| (!(pcm_s->stat.nint % (100*MAX_NSAMPLES/NSAMPLES))))
		if (!handler_codec_status(pcm_s))
			if (pcm_s->on_rt_error)
				pcm_s->stat.nrt_errors++;
#endif

	/* Get timing data */
	pcm_s->stat.time_service = time_diff (&tv_begin1, &tv_end);

	return 0;


      __problems:
	snd_printk ("pcm_studio: snd_studio_interrupt: Error Codec\n");
	pcm_s->error_codec = 1;
	snd_codec_studio_stop(pcm_s->codec, pcm_s->ncodecs);
	return 1;
}


/***************************************************************/
/* This function updates "bitmap_read" field of codecs in      */
/* order to mark which record channels must be read from       */
/* codec. The goal is reduce PCI reading needs.                */
/***************************************************************/
void snd_pcm_refresh_bitmaps_read (
	snd_pcm_studio_t *pcm_s)
{
	unsigned		icodec, isw;
	unsigned short		ich;
	snd_pcm_subchn_t	*subchn;
	snd_pcm1_subchn_t	*subchn1;

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		pcm_s->codec[icodec]->bitmap_read = 0x0000;

	/* Test opened record channels */
	for (subchn=pcm_s->pcm->capture.subchn; subchn; subchn=subchn->next)
		if ((subchn1=subchn->private_data))
			pcm_s->codec[subchn->number/16]->bitmap_read |=
				0x0001<<(subchn->number%16);

	/* Test active routings. See st_switch.c for routing conventions */
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	for (isw=0; isw<16; isw++){
		if ((ich = pcm_s->codec[icodec]->playback_routing[isw])
				> 16*pcm_s->ncodecs){
			ich -= 16*pcm_s->ncodecs + 1;
			pcm_s->codec[ich/16]->bitmap_read |= 0x0001<<(ich%16);
		}
		if ((ich = pcm_s->codec[icodec]->record_routing[isw])
				> 16*pcm_s->ncodecs){
			ich -= 16*pcm_s->ncodecs + 1;
			pcm_s->codec[ich/16]->bitmap_read |= 0x0001<<(ich%16);
		}
		if (isw > 1) continue;
		if ((ich = pcm_s->codec[icodec]->monitor_routing[isw])
				> 16*pcm_s->ncodecs){
			ich -= 16*pcm_s->ncodecs + 1;
			pcm_s->codec[ich/16]->bitmap_read |= 0x0001<<(ich%16);
		}
	}
}

/***************************************************************/
/* This boolean function tests if there are active routings.   */ 
/***************************************************************/
int snd_pcm_studio_test_routes(
	snd_pcm_studio_t *pcm_s)

{
	unsigned icodec;

	if (!pcm_s)
		return 0;

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if (snd_codec_studio_test_routes(pcm_s->codec[icodec]))
			return 1;

	return 0;
}

/***************************************************************/
/* This function manages the "usage" field variations due to   */
/* routing channels.                                           */
/***************************************************************/
void snd_pcm_studio_toggle_routing(
	snd_pcm_studio_t	*pcm_s,
	int			on_off)

{
	if (!pcm_s)
		return;

	/* Start codecs, if needed */
	if (on_off){
		if (pcm_s->usage == 0)
			handler_codec_open(pcm_s);
		pcm_s->usage++;
	}

	/* Stop codecs, if needed */
	if (!on_off){
		pcm_s->usage--;
		if (pcm_s->usage == 0)
			handler_codec_close(pcm_s);
	}
}

/***************************************************************/
void snd_pcm_studio_free (snd_pcm_studio_t *pcm_s)

{
	if (!pcm_s)
		return;

	if (timer_pending(&pcm_s->t_adat_sync))
		del_timer(&pcm_s->t_adat_sync);
	
	snd_pcm_studio_proc_done(pcm_s);

	if (pcm_s->bufmonitor)
		vfree(pcm_s->bufmonitor);
	if (pcm_s->bufrecord)
		vfree(pcm_s->bufrecord);
	if (pcm_s->bufplayback)
		vfree(pcm_s->bufplayback);

	/* Unregister pcm subsystem includes */
	/* "pcm" and "pcm1" free             */
	if (pcm_s->pcm)
		snd_pcm_unregister(pcm_s->pcm);

	snd_kfree(pcm_s);
}

/*****************************************************************/
snd_pcm_studio_t *snd_pcm_studio_create (
	snd_card_t	*card,
	unsigned	ipcm,
	unsigned	ncards)

{
	snd_pcm_t		*pcm;
	snd_pcm1_channel_t	*pchn1_p;
	snd_pcm1_channel_t	*pchn1_r;
	snd_pcm_studio_t	*pcm_s;
	unsigned long		bufsize;

	if (!ncards){
		snd_printk ("pcm_studio: snd_pcm_studio_create: ncards == 0\n");
		return NULL;
	}

	if (!(pcm = snd_pcm1_new_device(
			card, "Studi/o", 16*ncards, 16*ncards))){
		snd_printk ("pcm_studio: snd_pcm_studio_create: snd_pcm1_new_device() failed\n");
		return NULL;
	}

	pchn1_p = (snd_pcm1_channel_t *) pcm->playback.private_data;
	memcpy(&pchn1_p->hw, &snd_studio_playback, sizeof(snd_studio_playback));
	pchn1_p->private_data = NULL;

	pchn1_r = (snd_pcm1_channel_t *) pcm->capture.private_data;
	memcpy(&pchn1_r->hw, &snd_studio_record, sizeof(snd_studio_record));
	pchn1_r->private_data = NULL;

	pcm->info_flags	= SND_PCM_INFO_PLAYBACK
			| SND_PCM_INFO_CAPTURE
			| SND_PCM_INFO_DUPLEX;
	strcpy(pcm->name, "Studi/o");

	if (snd_pcm_register(pcm, 0) < 0) {
		snd_pcm_free(pcm); /* snd_pcm_free() frees pcm1 extension too */
		snd_printk ("pcm_studio: snd_pcm_studio_create: snd_pcm_register() failed\n");
		return NULL;
	}

	/* Our pcm object */
	if (!(pcm_s = (snd_pcm_studio_t *) snd_kcalloc(sizeof(*pcm_s), GFP_KERNEL))){
		snd_pcm_unregister(pcm); /* snd_pcm_unregister includes snd_pcm_free */
		snd_printk ("pcm_studio: snd_pcm_studio_create: snd_kcalloc() failed\n");
		return NULL;
	}

	init_timer(&pcm_s->t_adat_sync);
	pcm_s->access_lock		= SPIN_LOCK_UNLOCKED;
	pcm_s->card			= card;
	pcm_s->pcm			= pcm;
	pcm_s->ipcm			= ipcm;
	pcm_s->frame_per_channel	= snd_codec_get_channel_size();
	pcm_s->frame_per_codec		= 16*pcm_s->frame_per_channel;

	/* Playback buffer */
	bufsize  = pcm_s->frame_per_codec*ncards;
	if (!(pcm_s->bufplayback = vmalloc(bufsize))){
		snd_printk ("pcm_studio: snd_pcm_studio_create: Problems allocating playback buffer\n");
		goto __problems;
		}

	/* Record buffer */
	bufsize  = pcm_s->frame_per_codec*ncards;
	if (!(pcm_s->bufrecord = vmalloc(bufsize))){
		snd_printk ("pcm_studio: snd_pcm_studio_create: Problems allocating record buffer\n");
		goto __problems;
	}

	/* Monitor buffer */
	bufsize  = 2*pcm_s->frame_per_channel*ncards;
	if (!(pcm_s->bufmonitor = vmalloc(bufsize))){
		snd_printk ("pcm_studio: snd_pcm_studio_create: Problems allocating monitor buffer\n");
		goto __problems;
	}


	/* "private_data" argument for functions in */
	/* "snd_stru_pcm1_hardware" structures      */
	pchn1_p->private_data = pcm_s;
	pchn1_r->private_data = pcm_s;

	if (snd_pcm_studio_proc_init(pcm_s)){
		snd_printk ("pcm_studio: snd_pcm_studio_create: snd_pcm_studio_proc_init() failed\n");
		goto __problems;
	}

	return pcm_s;

      __problems:
	snd_pcm_studio_free (pcm_s);
	return NULL;
}
