#include "studio.h"
#include "codec_studio.h"
#include "pcm_studio.h"

/***************************************************************/
/* Alsa switches for Studi/o soundcards                        */
/* 	Routing switches                                       */
/* 	Meter switches                                      */
/***************************************************************/

/***************************************************************/
/* ROUTING. This layer manages data to route data among        */
/* channels at driver level. Routing data is local to codec.   */
/* However, it is managed by pcm layer. Therefore, routing is  */
/* posible among channels of different soundcards.             */
/*                                                             */
/* Data transfer to/from user level is relied on Alsa switches */
/* at control switch list. One master card manages all routing */
/* of its codecs from its control file descriptor. For each    */
/* codec (i.e. soundcard) there are three destination switches:*/
/* Playback, Record and Monitor switch, giving user access to  */
/* three destination routing tables: Playback (16 words),      */
/* Record (16 words) and Monitor (2 words).                    */
/*                                                             */
/* Each destination switch contains the index of the source    */
/* channel to be routed "TO". The index of the source channels */
/* is offset by the index card. First all playback sources,    */
/* and then all record sources. Example for two soundcards:    */
/*                                                             */
/*     Source Playback channel  #1 - card #0 ....... :  1      */
/*     Source Playback channel  #2 - card #0 ....... :  2      */
/*     Source Playback channel #16 - card #0 ....... : 16      */
/*                                                             */
/*     Source Playback channel  #1 - card #1 ....... : 17      */
/*     Source Playback channel  #2 - card #1 ....... : 18      */
/*     Source Playback channel #16 - card #1 ....... : 32      */
/*                                                             */
/*     Source Record   channel  #1 - card #0 ....... : 33      */
/*     Source Record   channel  #2 - card #0 ....... : 34      */
/*     Source Record   channel #16 - card #0 ....... : 48      */
/*                                                             */
/*     Source Record   channel  #1 - card #1 ....... : 49      */
/*     Source Record   channel  #2 - card #1 ....... : 50      */
/*     Source Record   channel #16 - card #1 ....... : 64      */
/*                                                             */
/* Index 0 is meant no routing                                 */
/***************************************************************/

#define SND_STUDIO_PLAYBACK_ROUTING	0
#define SND_STUDIO_RECORD_ROUTING	1
#define SND_STUDIO_MONITOR_ROUTING	2
#define SND_STUDIO_PLAYBACK0_METERS	3
#define SND_STUDIO_PLAYBACK1_METERS	4
#define SND_STUDIO_RECORD0_METERS	5
#define SND_STUDIO_RECORD1_METERS	6
#define SND_STUDIO_MONITOR_METERS	7

/***************************************************************/
static int snd_studio_get_switch(
		void		*_card,
		snd_kswitch_t	*kswitch,
		snd_switch_t	*uswitch)
{
	codec_studio_t		*codec;
	void			*table;
	unsigned		tbl_size;

	if (!(codec=(codec_studio_t *)kswitch->private_data))
		return -ENODEV;

	switch (kswitch->private_value){
		case SND_STUDIO_PLAYBACK_ROUTING:
			table = codec->playback_routing;
			tbl_size = sizeof(codec->playback_routing);
			break;
		case SND_STUDIO_RECORD_ROUTING:
			table = codec->record_routing;
			tbl_size = sizeof(codec->record_routing);
			break;
		case SND_STUDIO_MONITOR_ROUTING:
			table = codec->monitor_routing;
			tbl_size = sizeof(codec->monitor_routing);
			break;
		case SND_STUDIO_PLAYBACK0_METERS:
			table    = &codec->status.peak_playback[0];
			tbl_size = sizeof(codec->status.peak_playback)/2;
			break;
		case SND_STUDIO_PLAYBACK1_METERS:
			table    = &codec->status.peak_playback[8];
			tbl_size = sizeof(codec->status.peak_playback)/2;
			break;
		case SND_STUDIO_RECORD0_METERS:
			table    = &codec->status.peak_record[0];
			tbl_size = sizeof(codec->status.peak_record)/2;
			break;
		case SND_STUDIO_RECORD1_METERS:
			table    = &codec->status.peak_record[8];
			tbl_size = sizeof(codec->status.peak_record)/2;
			break;
		case SND_STUDIO_MONITOR_METERS:
			table    = &codec->status.peak_monitor[0];
			tbl_size = sizeof(codec->status.peak_monitor);
			break;
		default:
			return -EINVAL;
	}

	switch (kswitch->private_value){
		case SND_STUDIO_PLAYBACK_ROUTING:
		case SND_STUDIO_RECORD_ROUTING:
		case SND_STUDIO_MONITOR_ROUTING:
			uswitch->type = SND_SW_TYPE_WORD;
			memcpy (&uswitch->value.data16, table, tbl_size);
			break;
		case SND_STUDIO_PLAYBACK0_METERS:
		case SND_STUDIO_PLAYBACK1_METERS:
		case SND_STUDIO_RECORD0_METERS:
		case SND_STUDIO_RECORD1_METERS:
		case SND_STUDIO_MONITOR_METERS:
			uswitch->type = SND_SW_TYPE_DWORD;
			memcpy (&uswitch->value.data32, table, tbl_size);
			break;
		default:
			return -EINVAL;
	}

	return 0;
}

/***************************************************************/
/* If we are setting routing switches we must toggle codec     */
/* status (start/stop) if needed. This toggle control is       */
/* managed by pcm layer.                                       */
/***************************************************************/
static int snd_studio_set_switch (
		void		*_card,
		snd_kswitch_t	*kswitch,
		snd_switch_t	*uswitch)
{
	unsigned long		flags;
	snd_pcm_studio_t	*pcm_s;
	codec_studio_t		*codec;
	unsigned short		*table;
	unsigned		ntbl, idx;
	int			is_routing    = 0;
	int			route_toggle0 = 0;
	int			route_toggle1 = 0;

	if (uswitch->type != SND_SW_TYPE_WORD)
		return -EINVAL;

	if (!(codec=(codec_studio_t *)kswitch->private_data))
		return -ENODEV;
	pcm_s = codec->pcm_s;

	switch (kswitch->private_value){
		case SND_STUDIO_PLAYBACK_ROUTING:
			table = &codec->playback_routing[0];
			ntbl = 16;
			break;
		case SND_STUDIO_RECORD_ROUTING:
			table = &codec->record_routing[0];
			ntbl = 16;
			break;
		case SND_STUDIO_MONITOR_ROUTING:
			table = &codec->monitor_routing[0];
			ntbl = 2;
			break;
		default:
			return -EINVAL;
	}
	if (!table)
		return -ENODEV;

	/* Test range value */
	switch (kswitch->private_value){
		case SND_STUDIO_PLAYBACK_ROUTING:
		case SND_STUDIO_RECORD_ROUTING:
		case SND_STUDIO_MONITOR_ROUTING:
			is_routing = 1;
			for (idx=0; idx<ntbl; idx++)
			if (uswitch->value.data16[idx] > 32*pcm_s->ncodecs)
				return -EINVAL;
			break;
	}

	if (is_routing)
		route_toggle0 = snd_pcm_studio_test_routes(pcm_s);

	spin_lock_irqsave(&pcm_s->access_lock, flags);
	memcpy (table, &uswitch->value.data16[0],
			ntbl*sizeof(uswitch->value.data16[0]));
	if (is_routing)
		snd_pcm_refresh_bitmaps_read (pcm_s);
	spin_unlock_irqrestore(&pcm_s->access_lock, flags);

	if (is_routing){
		route_toggle1 = snd_pcm_studio_test_routes(pcm_s);
		if (route_toggle0 != route_toggle1)
			snd_pcm_studio_toggle_routing(
				pcm_s, route_toggle1);
	}

	return 0;
}

/*********************************************************/
static int _snd_studio_free_switch (
	snd_card_t		*card,
	codec_studio_t		*codec,
	unsigned		type)
{
	snd_kswitch_t	**ppswitch;
	int		error;

	switch (type){
		case SND_STUDIO_PLAYBACK_ROUTING:
			ppswitch = &codec->sw_playback_routing;
			break;
		case SND_STUDIO_RECORD_ROUTING:
			ppswitch = &codec->sw_record_routing;
			break;
		case SND_STUDIO_MONITOR_ROUTING:
			ppswitch = &codec->sw_monitor_routing;
			break;
		case SND_STUDIO_PLAYBACK0_METERS:
			ppswitch = &codec->sw_playback0_meters;
			break;
		case SND_STUDIO_PLAYBACK1_METERS:
			ppswitch = &codec->sw_playback1_meters;
			break;
		case SND_STUDIO_RECORD0_METERS:
			ppswitch = &codec->sw_record0_meters;
			break;
		case SND_STUDIO_RECORD1_METERS:
			ppswitch = &codec->sw_record1_meters;
			break;
		case SND_STUDIO_MONITOR_METERS:
			ppswitch = &codec->sw_monitor_meters;
			break;
		default:
			return -EINVAL;
	}

	/* If it is already removed, return without error */
	if (*ppswitch)
		if (!(error=snd_control_switch_remove(card, *ppswitch)))
			return error;
	*ppswitch = NULL;

	return 0;
}

/*********************************************************/
int snd_studio_free_switches (
	snd_card_t		*card,
	snd_pcm_studio_t	*pcm_s)
{
	int error, icodec;

	if (!card)
		return -ENODEV;
	if (!pcm_s)
		return -ENODEV;

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if ((error=_snd_studio_free_switch (
				card, pcm_s->codec[icodec],
				SND_STUDIO_PLAYBACK_ROUTING)))
			return error;
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if ((error=_snd_studio_free_switch (
				card, pcm_s->codec[icodec],
				SND_STUDIO_RECORD_ROUTING)))
			return error;
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if ((error=_snd_studio_free_switch (
				card, pcm_s->codec[icodec],
				SND_STUDIO_MONITOR_ROUTING)))
			return error;
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if ((error=_snd_studio_free_switch (
				card, pcm_s->codec[icodec],
				SND_STUDIO_PLAYBACK0_METERS)))
			return error;
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if ((error=_snd_studio_free_switch (
				card, pcm_s->codec[icodec],
				SND_STUDIO_PLAYBACK1_METERS)))
			return error;
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if ((error=_snd_studio_free_switch (
				card, pcm_s->codec[icodec],
				SND_STUDIO_RECORD0_METERS)))
			return error;
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if ((error=_snd_studio_free_switch (
				card, pcm_s->codec[icodec],
				SND_STUDIO_RECORD1_METERS)))
			return error;
	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
		if ((error=_snd_studio_free_switch (
				card, pcm_s->codec[icodec],
				SND_STUDIO_MONITOR_METERS)))
			return error;

	return 0;
}


/*********************************************************/
static int _snd_studio_create_switch (
	snd_card_t		*card,
	unsigned		icard,	/* [0, ncards-1] */
	codec_studio_t		*codec,
	unsigned		icodec,	/* [0, ncodecs-1] */
	unsigned		type)
{
	snd_kswitch_t	kswitch;
	snd_kswitch_t 	**ppswitch;

	kswitch.get		= snd_studio_get_switch;
	kswitch.set		= snd_studio_set_switch;
	kswitch.private_data	= NULL;
	kswitch.private_free	= NULL;
	kswitch.private_value	= type;

	switch (type){
		case SND_STUDIO_PLAYBACK_ROUTING:
			ppswitch = &codec->sw_playback_routing;
			sprintf (kswitch.name, "Play Routes (%d-%d) of Card#%u",
					icodec*16+1, icodec*16+16, icard);
			break;
		case SND_STUDIO_RECORD_ROUTING:
			ppswitch = &codec->sw_record_routing;
			sprintf (kswitch.name, "Rec Routes (%d-%d) of Card#%u",
					icodec*16+1, icodec*16+16, icard);
			break;
		case SND_STUDIO_MONITOR_ROUTING:
			ppswitch = &codec->sw_monitor_routing;
			sprintf (kswitch.name, "Mon Routes (%d-%d) of Card#%u",
					icodec*2+1, icodec*2+2, icard);
			break;
		case SND_STUDIO_PLAYBACK0_METERS:
			ppswitch = &codec->sw_playback0_meters;
			sprintf (kswitch.name, "Play Meters (%d-%d) of Card#%u",
					icodec*16+1, icodec*16+8, icard);
			break;
		case SND_STUDIO_PLAYBACK1_METERS:
			ppswitch = &codec->sw_playback1_meters;
			sprintf (kswitch.name, "Play Meters (%d-%d) of Card#%u",
					icodec*16+9, icodec*16+16, icard);
			break;
		case SND_STUDIO_RECORD0_METERS:
			ppswitch = &codec->sw_record0_meters;
			sprintf (kswitch.name, "Rec Meters (%d-%d) of Card#%u",
					icodec*16+1, icodec*16+8, icard);
			break;
		case SND_STUDIO_RECORD1_METERS:
			ppswitch = &codec->sw_record1_meters;
			sprintf (kswitch.name, "Rec Meters (%d-%d) of Card#%u",
					icodec*16+9, icodec*16+16, icard);
			break;
		case SND_STUDIO_MONITOR_METERS:
			ppswitch = &codec->sw_monitor_meters;
			sprintf (kswitch.name, "Mon Meters (%d-%d) of Card#%u",
					icodec*2+1, icodec*2+2, icard);
			break;
		default:
			return -EINVAL;
	}

	/* If it is already created, return without error */
	if (*ppswitch != NULL)
		return 0;

        if (!(*ppswitch = snd_control_switch_new(
				card, &kswitch, codec)))
		return -ENOMEM;

	return 0;
}

/*********************************************************/
int snd_studio_create_switches (
	snd_card_t		*card,
	unsigned		icard,
	snd_pcm_studio_t	*pcm_s)
{
	int	error, icodec;

	if (!card)
		return -ENODEV;
	if (!pcm_s)
		return -ENODEV;

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((error=_snd_studio_create_switch (
				card, icard, pcm_s->codec[icodec], icodec,
				SND_STUDIO_PLAYBACK_ROUTING))){
			snd_studio_free_switches (card, pcm_s);
			return error;
	}

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((error=_snd_studio_create_switch (
				card, icard, pcm_s->codec[icodec], icodec,
				SND_STUDIO_RECORD_ROUTING))){
			snd_studio_free_switches (card, pcm_s);
			return error;
	}

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((error=_snd_studio_create_switch (
				card, icard, pcm_s->codec[icodec], icodec,
				SND_STUDIO_MONITOR_ROUTING))){
			snd_studio_free_switches (card, pcm_s);
			return error;
	}

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((error=_snd_studio_create_switch (
				card, icard, pcm_s->codec[icodec], icodec,
				SND_STUDIO_PLAYBACK0_METERS))){
			snd_studio_free_switches (card, pcm_s);
			return error;
	}

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((error=_snd_studio_create_switch (
				card, icard, pcm_s->codec[icodec], icodec,
				SND_STUDIO_PLAYBACK1_METERS))){
			snd_studio_free_switches (card, pcm_s);
			return error;
	}

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((error=_snd_studio_create_switch (
				card, icard, pcm_s->codec[icodec], icodec,
				SND_STUDIO_RECORD0_METERS))){
			snd_studio_free_switches (card, pcm_s);
			return error;
	}

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((error=_snd_studio_create_switch (
				card, icard, pcm_s->codec[icodec], icodec,
				SND_STUDIO_RECORD1_METERS))){
			snd_studio_free_switches (card, pcm_s);
			return error;
	}

	for (icodec=0; icodec<pcm_s->ncodecs; icodec++)
	if ((error=_snd_studio_create_switch (
				card, icard, pcm_s->codec[icodec], icodec,
				SND_STUDIO_MONITOR_METERS))){
			snd_studio_free_switches (card, pcm_s);
			return error;
	}

	return 0;
}
